// UC: 21020 - Computação Gráfica (2022/23)
// Aluno: 1902551 - Pedro Gaspar

import * as THREE from "https://unpkg.com/three@0.124.0/build/three.module.js";
import { OrbitControls } from "https://unpkg.com/three@0.124.0/examples/jsm/controls/OrbitControls.js";
import { BezierCurve } from "./bezier-curve.mjs";
import { randomFloat, throttle, round } from "./helpers.mjs";

// /////////////////////////////////////////////////////////////////////////////
// #region Variáveis globais

let screenWidth = window.innerWidth;
let screenHeight = window.innerHeight;

/**
 * Posição do rato relativo ao topo esquerdo da janela
 */
const MOUSE_WINDOW_POSITION = { x: 0, y: 0 };
/**
 * Posição do rato relativo ao centro do tabuleiro
 */
const MOUSE_PLANE_POSITION = { x: 0, y: 0 };

const SPHERES_INFO_BOX = document.getElementById("spheres-info");
const MOUSE_INFO_BOX = document.getElementById("mouse-info");

const TILE_WIDTH = 1;
const TILES_PER_ROW = 20;
const GRID_WIDTH = TILES_PER_ROW * TILE_WIDTH;

/**
 * Número de casas decimais (precisão) a utilizar nos cálculos
 */
const PRECISION = 3;

/** Valor que uma bola sobe/desce de cada vez */
const Z_STEP = 0.1;

/**
 * Valor máximo que uma bola pode subir/descer,
 * se o utilizador mantiver a tecla de subir ou descer pressionada
 * */
const MAX_ABSOLUTE_Z_STEP = 1.5;

const CAMERA = new THREE.PerspectiveCamera(
  45, // fov
  screenWidth / screenHeight, // aspect
  1, // near
  10000 // far
);

const SCENE = new THREE.Scene();
const RENDERER = new THREE.WebGLRenderer({ antialias: true });
const CONTROLS = new OrbitControls(CAMERA, RENDERER.domElement);
const RAYCASTER = new THREE.Raycaster();
const POINTER = new THREE.Vector2();

let selectedSphereIndex = -1;
let keyRepetitions;

const SCENE_OBJECTS = [];
let SPHERES = [];

/**
 * Inicialização do programa, desenha o tabuleiro e coloca as bolas
 */
function init() {
  CAMERA.position.set(0, 0, GRID_WIDTH * 1.5);
  CAMERA.lookAt(0, 0, 0);

  SCENE.background = new THREE.Color("#fff");

  CONTROLS.minPolarAngle = 0;

  RENDERER.setSize(screenWidth, screenHeight);
  RENDERER.setPixelRatio(window.devicePixelRatio);
  document.body.appendChild(RENDERER.domElement);

  drawGrid();
  initSpheres();

  // Aqui uso o throttle para prevenir que o evento seja processado demasiadas
  // vezes. Assim, só ocorre uma vez a cada 25 ms, o que corresponde a 40 FPS.
  document.addEventListener("pointermove", throttle(onPointerMove, 25));
  document.addEventListener("keydown", onKeydown);
  window.addEventListener("resize", onResize);

  // Caixas de informação
  setSpheresInfoBox();
  setMouseInfoBox();
}

/**
 * Desenha o tabuleiro, assim como as linhas dos eixos X, Y e Z
 */
function drawGrid() {
  const grid = new THREE.PlaneGeometry();

  const tileIndices = [...Array(TILES_PER_ROW).keys()].map(
    (x) => (x -= Math.floor(TILES_PER_ROW / 2))
  );

  for (const row of tileIndices) {
    for (const col of tileIndices) {
      const pair = (row + col) % 2;

      const tile = new THREE.Mesh(
        grid,
        new THREE.LineBasicMaterial({
          color: pair ? "#fff" : "#888",
          linewidth: 1,
          side: THREE.DoubleSide,
          transparent: true,
          opacity: pair ? 0.1 : 0.25,
        })
      );

      // O nome é utilizado para detectar a interseção do rato com os ladrilhos
      tile.name = "tile";
      tile.position.set(
        row * TILE_WIDTH + TILE_WIDTH / 2,
        col * TILE_WIDTH + TILE_WIDTH / 2,
        0
      );
      // Rodar 90 graus
      SCENE.add(tile);
    }
  }

  // Linha das ordenadas (y) (vermelha)
  const yLineCoords = [
    new THREE.Vector3(0, 0, 0),
    new THREE.Vector3(0, GRID_WIDTH / 2, 0),
  ];
  const yLine = new THREE.Line(
    new THREE.BufferGeometry().setFromPoints(yLineCoords),
    new THREE.LineBasicMaterial({ color: "#f00" })
  );
  SCENE.add(yLine);

  // Linha das abcissas (x) (azul)
  const xLineCoords = [
    new THREE.Vector3(0, 0, 0),
    new THREE.Vector3(GRID_WIDTH / 2, 0, 0),
  ];
  const xLine = new THREE.Line(
    new THREE.BufferGeometry().setFromPoints(xLineCoords),
    new THREE.LineBasicMaterial({ color: "#00f" })
  );
  SCENE.add(xLine);

  // Linha das cotas (z) (verde)
  const zLineCoords = [
    new THREE.Vector3(0, 0, -GRID_WIDTH / 2),
    new THREE.Vector3(0, 0, GRID_WIDTH / 2),
  ];
  const zLine = new THREE.Line(
    new THREE.BufferGeometry().setFromPoints(zLineCoords),
    new THREE.LineBasicMaterial({ color: "#0f0" })
  );
  SCENE.add(zLine);
}

/**
 * Desenha as bolas no plano, em posições aleatórias
 */
function initSpheres() {
  const sphereColors = ["yellow", "orange", "red", "green", "blue"];

  for (const color of sphereColors) {
    const sphere = new THREE.Mesh(
      new THREE.SphereGeometry(0.5, 16, 8),
      new THREE.MeshBasicMaterial({
        color: new THREE.Color(color),
        transparent: true,
        opacity: 0.4,
      })
    );

    // Coloca a bola em coordenadas X,Y aleatórias
    sphere.position.set(
      randomFloat(-GRID_WIDTH / 2, GRID_WIDTH / 2, PRECISION),
      randomFloat(-GRID_WIDTH / 2, GRID_WIDTH / 2, PRECISION),
      0
    );

    SPHERES.push(sphere);
    SCENE.add(sphere);
  }
}

/**
 * Remove a seleção das bolas, colocando-as semi-transparentes
 */
function deselectSpheres() {
  SPHERES.forEach((sphere) => {
    sphere.material.opacity = 0.4;
  });

  SCENE.children
    .filter((obj) => obj.name.startsWith("sphereZLine"))
    .forEach((obj) => {
      obj.material.opacity = 0.4;
    });
}

/**
 * Selecciona a bola na posição indicada.
 * Se a bola já estiver seleccionada, remove a sua seleção.
 */
function selectSphere(position) {
  deselectSpheres();

  // Se carregar de novo na mesma bola, deselciona-a
  if (selectedSphereIndex === position) {
    selectedSphereIndex = -1;
    setSpheresInfoBox();
    return;
  }

  selectedSphereIndex = position;
  SPHERES[selectedSphereIndex].material.opacity = 1;

  SCENE.children
    .filter((obj) => obj.name === `sphereZLine${selectedSphereIndex}`)
    .forEach((obj) => {
      obj.material.opacity = 1;
    });

  setSpheresInfoBox();
}

/**
 * Posiciona a bola selecionada no plano, de acordo com a posição do rato
 */
function setSpherePosition() {
  if (selectedSphereIndex < 0) {
    return;
  }

  SPHERES[selectedSphereIndex].position.x = MOUSE_PLANE_POSITION.x;
  SPHERES[selectedSphereIndex].position.y = MOUSE_PLANE_POSITION.y;

  setSphereZLine(selectedSphereIndex);
  setSpheresInfoBox();
}

/**
 * Move a bola selecionada para cima ou para baixo no eixo Z
 * @param {"up" | "down"} direction Direção do movimento (cima ou baixo)
 * @param {number} steps Quantidade de passos de 0.1 a dar.
 * Será multiplicado por 0.1 para obter a distância a mover.
 */
function moveSphereZ(direction, steps = 1) {
  if (selectedSphereIndex < 0) {
    return;
  }

  if (direction !== "up" && direction !== "down") {
    throw new Error(
      "Direção do movimento no eixo Z inválida. Os valores só podem ser 'up' ou 'down'."
    );
  }

  let amount = steps * Z_STEP;

  // Previne que a bola comece a subir em intervalos demasiado grandes ou
  // demasiado pequenos
  if (amount > MAX_ABSOLUTE_Z_STEP) {
    amount = MAX_ABSOLUTE_Z_STEP;
  } else if (amount < Z_STEP) {
    amount = Z_STEP;
  }

  if (direction === "down") {
    amount = -amount;
  }

  SPHERES[selectedSphereIndex].position.z += amount;

  // Desenha a linha do eixo Z entre o centro da bola e a origem do eixo,
  // se necessário
  setSphereZLine();

  setSpheresInfoBox();
}

/**
 * Desenha uma linha reta entre o centro da bola selecionada e a origem do eixo.
 * Se não houver bola selecionada ou se a bola estiver na origem, não faz nada.
 */
function setSphereZLine() {
  if (selectedSphereIndex < 0) {
    return;
  }

  // Remove a linha anterior da bola, se existir
  SCENE.children
    .filter((child) => child.name === "sphereZLine" + selectedSphereIndex)
    .forEach((child) => {
      SCENE.remove(child);
    });

  const { x, y, z } = SPHERES[selectedSphereIndex].position;

  // A bola está centrada no eixo Z, não precisamos de desenhar
  if (z === 0) {
    return;
  }

  const zLine = new THREE.Line(
    new THREE.BufferGeometry().setFromPoints([
      new THREE.Vector3(x, y, 0),
      new THREE.Vector3(x, y, z),
    ]),
    new THREE.LineBasicMaterial({ color: "#000", transparent: true })
  );

  zLine.name = `sphereZLine${selectedSphereIndex}`;

  SCENE.add(zLine);
  SCENE_OBJECTS.push(zLine);
}

/**
 * Desenha a curva de Bézier entre as bolas, atribuindo-lhe uma cor aleatória
 */
function drawBezierCurve() {
  // Inicializa a THREE.Curve com a nossa implementação, que calcula a curva de
  // Bézier utilizando a função bezier4.
  const curve = new BezierCurve(SPHERES.map((sphere) => sphere.position));

  // https://css-tricks.com/snippets/javascript/random-hex-color/
  const curveColor = `#${Math.floor(Math.random() * 16777215).toString(16)}`;

  // A curva fica com uma grossura de 2/3 da bola
  const mesh = new THREE.Mesh(
    new THREE.TubeGeometry(curve, 20, (TILE_WIDTH * 0.66) / 2, 8, false),
    new THREE.MeshBasicMaterial({
      color: new THREE.Color(curveColor),
      side: THREE.DoubleSide,
    })
  );

  SCENE.add(mesh);
  SCENE_OBJECTS.push(mesh);

  selectedSphereIndex = -1;
  deselectSpheres();
  setSpheresInfoBox();
}

/**
 * Limpa a cena, removendo todas as bolas e curvas. Volta a colocar as bolas
 * em posições aleatórias no plano.
 */
function clearScene() {
  selectedSphereIndex = -1;
  SCENE_OBJECTS.forEach((object) => SCENE.remove(object));
  SPHERES.forEach((sphere) => SCENE.remove(sphere));
  SPHERES = [];

  initSpheres();
}

/**
 * Chamado ao pressionar uma tecla no teclado. Trata de chamar as funções
 * correspondentes aos controlos do plano. Tem lógica para gravar se uma tecla
 * se mantém pressionada, e quantas repetições ocorreram.
 * @param {KeyboardEvent} event Evento do teclado utilizado para saber a tecla
 * pressionada, e se é uma repetição
 */
function onKeydown(event) {
  const { key, code, repeat } = event;

  // Se o repeat estiver a true, significa que o utilizador pressionou e manteve
  // pressionada a tecla. Gravamos quantas vezes é pressionada, para incrementar
  // a distância a mover a bola no eixo Z.
  if (repeat) {
    if (keyRepetitions != null && keyRepetitions.key === code) {
      keyRepetitions.count += 1;
    } else {
      keyRepetitions = { key: code, count: 1 };
    }
  } else {
    keyRepetitions = null;
  }

  const number = parseInt(key);
  if (!isNaN(number) && number <= SPHERES.length) {
    return void selectSphere(number - 1);
  }

  switch (code) {
    case "Backspace": // Limpar cena
      clearScene();
      break;
    case "Space": // Posicionar a bola onde o rato está
      setSpherePosition();
      break;
    case "KeyX": // Desenhar a curva
      drawBezierCurve();
      setSpheresInfoBox();
      break;
    case "KeyW": // Mover a bola selecionada para cima ou para baixo
    case "KeyS": {
      const step = keyRepetitions?.key === code ? keyRepetitions.count : 1;
      moveSphereZ(code == "KeyW" ? "up" : "down", step);
      break;
    }
  }
}

/**
 * Chamado quando o rato é movido na janela. Grava a posição do rato na janela,
 * e calcula a posição relativo ao centro do plano, para que quando o utilizador
 * pressione o espaço, a bola seja movida para essas coordenadas.
 * @param {MouseEvent} event
 */
function onPointerMove(event) {
  MOUSE_WINDOW_POSITION.x = event.clientX;
  MOUSE_WINDOW_POSITION.y = event.clientY;

  // O pointer é um simples vetor dummy que utilizamos para verificar se
  // interceta com alguma telha do plano. Este é fornecido ao raycaster que
  // por sua vez verifica se existe alguma interceção com os elementos do plano.
  POINTER.set(
    round((event.clientX / screenWidth) * 2 - 1, PRECISION),
    round(-(event.clientY / screenHeight) * 2 + 1, PRECISION)
  );
  RAYCASTER.setFromCamera(POINTER, CAMERA);

  const intersections = RAYCASTER.intersectObjects(
    SCENE.children.filter((child) => child.name === "tile"),
    false
  );

  // Se houver interseção, utilizamos o ponto onde esta ocorre para gravar
  // a posição do rato relativamente ao centro do plano.
  if (intersections.length) {
    const [intersection] = intersections;

    if (intersection.face.normal) {
      const { x, y } = intersection.point;

      MOUSE_PLANE_POSITION.x = round(x, PRECISION);
      MOUSE_PLANE_POSITION.y = round(y, PRECISION);
    }
  }

  setMouseInfoBox();
}

/**
 * Chamado quando o tamanho da janela altera. Atualiza o aspect ratio da câmara
 * e do renderer para manter o canvas sempre a ocupar a janela inteira.
 */
function onResize() {
  screenWidth = window.innerWidth;
  screenHeight = window.innerHeight;

  CAMERA.aspect = screenWidth / screenHeight;
  CAMERA.updateProjectionMatrix();

  RENDERER.setSize(screenWidth, screenHeight);
}

/**
 * Inicia a renderização do canvas.
 */
function render() {
  requestAnimationFrame(render);
  CONTROLS.update();
  RENDERER.render(SCENE, CAMERA);
}

/**
 * Atualiza os valores da caixa de informação das bolas.
 * A caixa indica as coordenadas x,y,z de cada bola, a sua cor, e se está
 * selecionada atualmente.
 */
function setSpheresInfoBox() {
  SPHERES_INFO_BOX.innerHTML = "";

  let i = 0;
  for (const sphere of SPHERES) {
    const { r, g, b } = sphere.material.color;

    SPHERES_INFO_BOX.innerHTML += `
      <div class="sphere ${i === selectedSphereIndex ? "selected" : ""}">
        <div
          class="circle"
          style="background-color: rgb(${r * 255}, ${g * 255}, ${b * 255});"
        ><span>${i + 1}</span></div>

        <code>
          {
            x: ${sphere.position.x},
            y: ${sphere.position.y},
            z: ${round(sphere.position.z, 3)}
          }
        </code>
      </div>`;

    i++;
  }
}

/**
 * Atualiza os valores da caixa de informações do rato.
 * A caixa indica a posição atual do rato em pixeis em relação ao topo esquerdo
 * da janela, e em relação ao centro do plano.
 */
function setMouseInfoBox() {
  MOUSE_INFO_BOX.innerHTML = `
    <p>
      Janela
      <code>
        {
          x: ${MOUSE_WINDOW_POSITION.x},
          y: ${MOUSE_WINDOW_POSITION.y}
        }
      </code>
    </p>
    <p>
      Plano
      <code>
        {
          x: ${MOUSE_PLANE_POSITION.x},
          y: ${MOUSE_PLANE_POSITION.y}
        }
      </code>
    </p>
  `;
}

init();
render();
